from me_main_libs import *

from bokeh.palettes import Dark2_5, Category20c, brewer
from bokeh.transform import cumsum
from bokeh.io import output_file
from bokeh.plotting import figure
from bokeh.palettes import brewer
from bokeh.models import HoverTool, DatetimeTickFormatter, NumeralTickFormatter

import itertools
from math import pi


def _create_colors():
    colors = itertools.cycle(Category20c)
    return colors


def plot_bar_basic(df, x_name, y_name, title, plot_height=350, plot_width=None,
                   output_file_=None):
    '''
    output_file:'*.html'
    '''
    if output_file_:
        output_file(output_file_)

    p = figure(x_range=df[x_name],
               plot_height=plot_height,
               plot_width=plot_width,
               title=title,
               toolbar_location=None,
               tools="hover",
               tooltips=f"@{x_name}: @{y_name}",
               )

    p.vbar(x=x_name, top=y_name, width=0.9, source=df)
    p.xgrid.grid_line_color = None

    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p


def plot_bar_stacked(df, x_name, y_names, title, y_range_start=0, y_range_end=None,
                     plot_height=350, plot_width=None,
                     output_file_=None):
    '''
    only works if len(y_names)>=3

    output_file:'*.html'
    '''

    if output_file_:
        output_file(output_file_)

    tooltips = []
    tooltips.append((f'{x_name}', f'@{x_name}'))
    for y_name in y_names:
        tooltips.append((f'{y_name}', f'@{y_name}'))

    p = figure(x_range=df[x_name], plot_height=plot_height,
               plot_width=plot_width,
               title=title,
               toolbar_location=None,
               tools="hover",
               tooltips=tooltips,
               )

    p.vbar_stack(y_names, x=x_name, width=0.9, color=brewer['Spectral'][len(y_names)],
                 source=df,
                 legend_label=y_names)

    p.x_range.range_padding = 0.05
    p.axis.minor_tick_line_color = None
    p.legend.orientation = "horizontal"
    p.legend.location = "top_left"

    p.xgrid.grid_line_color = None
    p.y_range.start = y_range_start
    if y_range_end:
        p.y_range.end = y_range_end

    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p


def plot_histogram(series, title, draw_curves=False,
                   plot_height=350, plot_width=None,
                   output_file_=None):

    if output_file_:
        output_file(output_file_)

    series = series.values.reshape(1, -1)[0]
    bins = 50
    hist, edges = np.histogram(series, density=True, bins=bins)

    p = figure(title=title, plot_height=plot_height, plot_width=plot_width,
               toolbar_location=None,
               )

    p.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:],
           fill_color="navy", line_color="white", alpha=0.5)

    if draw_curves:
        mu = series.mean()
        sigma = series.std()
        x = np.linspace(min(series), max(series), 100)
        pdf = 1/(sigma * np.sqrt(2*np.pi)) * np.exp(-(x-mu)**2 / (2*sigma**2))
        # cdf = ((1+scipy.special.erf((x-mu)/np.sqrt(2*sigma**2)))/2)*100
        p.line(x, pdf, line_color="#ff8888", line_width=4,
               alpha=0.7, legend_label="PDF")
        # p.line(x, cdf, line_color="orange", line_width=2,
        #        alpha=0.7, legend_label="CDF")

        p.legend.location = "center_right"

    p.yaxis.axis_label = 'Prob(x)'
    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p


def plot_candlestick(ohlc, title, plot_height=350, plot_width=None,
                     output_file_=None):
    '''
    https://docs.bokeh.org/en/latest/docs/gallery/candlestick.html
    '''
    if output_file_:
        output_file(output_file_)

    ohlc.index.name = 'date'
    ohlc.reset_index(inplace=True)

    inc = ohlc.close > ohlc.open
    dec = ohlc.open > ohlc.close
    w = 12*60*60*1000  # half day in ms

    tools = "crosshair,box_zoom,reset,save"

    p = figure(title=title, plot_height=plot_height, plot_width=plot_width,
               toolbar_location='right',
               tools=tools,
               x_axis_type="datetime"
               )
    p.xaxis.major_label_orientation = pi/4
    p.grid.grid_line_alpha = 0.3

    p1 = p.segment('date', 'high', 'date', 'low', color="black", source=ohlc)
    p.vbar(ohlc.date[inc], w, ohlc.open[inc], ohlc.close[inc],
           fill_color="#D5E1DD", line_color="black")
    p.vbar(ohlc.date[dec], w, ohlc.open[dec], ohlc.close[dec],
           fill_color="#F2583E", line_color="black")

    tooltips = []
    tooltips.append(('date', '@date{%F}'))
    for y_name in ['open', 'high', 'low', 'close']:
        tooltips.append((f'{y_name}', f'@{y_name}'))

    p.add_tools(HoverTool(tooltips=tooltips,
                          formatters={'@date': 'datetime'},
                          renderers=[p1],
                          mode='vline'))
    p.yaxis.formatter = NumeralTickFormatter(format="$0.00")
    p.xaxis.formatter = DatetimeTickFormatter(months="%m-%d")

    p.x_range.range_padding = 0.02
    p.toolbar.autohide = True
    p.toolbar.logo = None
    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p


def plot_line(series, title, x_axis_type='datetime', plot_height=350, plot_width=None, output_file_=None):
    '''
    x_axis_type: 'linear', 'log', 'datetime', 'mercator'
    output_file:'*.html'
    '''

    if output_file_:
        output_file(output_file_)

    tools = "crosshair,box_zoom,reset,save"
    p = figure(title=title, plot_height=plot_height, plot_width=plot_width,
               toolbar_location='right',
               tools=tools,
               x_axis_type=x_axis_type)

    p.line(series.index, series.squeeze())

    p.x_range.range_padding = 0
    p.toolbar.autohide = True
    p.toolbar.logo = None
    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p


def plot_multi_line(df, title, x_axis_type='datetime', plot_height=350, plot_width=None,
                    legend_orientation="horizontal",
                    legend_location="top_left",
                    output_file_=None):
    '''
    x_axis_type: 'linear', 'log', 'datetime', 'mercator'
    '''

    if output_file_:
        output_file(output_file_)

    tools = "crosshair,box_zoom,reset,save"
    p = figure(title=title, plot_height=plot_height, plot_width=plot_width,
               toolbar_location='right', tools=tools,
               x_axis_type=x_axis_type)

    colors = itertools.cycle(Dark2_5)
    for col, color in zip(df.columns, colors):
        p.line(df.index, df[col], color=color, legend_label=col)

    p.legend.location = legend_location
    p.legend.orientation = legend_orientation
    p.legend.click_policy = "hide"

    p.x_range.range_padding = 0
    p.toolbar.autohide = True
    p.toolbar.logo = None
    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p


def plot_pie_chart(df, x_name, y_name, title, plot_height=350, plot_width=500,
                   output_file_=None):

    if output_file_:
        output_file(output_file_)

    df['angle'] = df[y_name]/df[y_name].sum() * 2*pi
    df['color'] = Category20c[len(df)]

    p = figure(plot_height=plot_height,
               plot_width=plot_width,
               title=title,
               toolbar_location=None,
               tools="hover",
               tooltips=f"@{x_name}: @{y_name}",
               x_range=(-0.5, 1.0)
               )

    p.wedge(x=0, y=1, radius=0.4,
            start_angle=cumsum('angle', include_zero=True), end_angle=cumsum('angle'),
            line_color="white", fill_color='color', legend_field=x_name, source=df)

    p.axis.axis_label = None
    p.axis.visible = False
    p.grid.grid_line_color = None

    return p


def plot_stacked_area(df, title, x_axis_type='linear', y_range_start=0,
                      y_range_end=None, plot_height=350, plot_width=None,
                      output_file_=None):
    '''
    hover tooltips have not been implemented for them yet 
    x_axis_type: 'linear', 'log', 'datetime', 'mercator'
    output_file:'*.html'
    '''

    if output_file_:
        output_file(output_file_)

    p = figure(plot_height=plot_height,
               plot_width=plot_width,
               title=title,
               toolbar_location=None,
               tools='crosshair',
               x_range=(0, len(df)-1),
               x_axis_type=x_axis_type)

    df.index.name = 'index'
    N = len(df.columns)
    p1 = p.varea_stack(stackers=list(df.columns), x='index',
                       color=brewer['Spectral'][N],
                       legend_label=list(df.columns), source=df)

    p.axis.minor_tick_line_color = None
    p.legend.orientation = 'horizontal'
    p.legend.location = 'top_left'
    p.legend.click_policy = "hide"

    p.y_range.start = 0
    if y_range_end:
        p.y_range.end = y_range_end
    if plot_width:
        p.plot_width = plot_width
    else:
        p.sizing_mode = "stretch_width"
    return p



from PIL import Image
import PyPDF2
import os
import pandas as pd


def merge_pdfs(pdfs, output_pdf, remove_pdfs=False):
    output_pdf = PyPDF2.PdfFileWriter()
    open_pdfs = []
    for _pdf in pdfs:
        try:
            _x = open(_pdf, 'rb')
            open_pdfs.append(_x)
            x = PyPDF2.PdfFileReader(_x)
            output_pdf.appendPagesFromReader(x)
            if (x.getNumPages() % 2) != 0:
                output_pdf.addBlankPage()
        except Exception as e:
            print(e)

    with open(output_pdf, 'wb') as out_pdf:
        output_pdf.write(out_pdf)

    for y in open_pdfs:
        y.close()

    if remove_pdfs:
        for _pdf in pdfs:
            os.remove(_pdf)


def create_pdf_from_pngs(pngs, output_pdf):
    _pngs = list(map(_modify_rgba_to_rgb, pngs))
    _pngs[0].save(output_pdf, "PDF", resolution=100,
                  save_all=True, append_images=_pngs[1:])
    [os.remove(png) for png in pngs]


def _modify_rgba_to_rgb(png):
    rgba = Image.open(png)
    # white background
    rgb = Image.new('RGB', rgba.size, (255, 255, 255))
    # paste using alpha channel as mask
    rgb.paste(rgba, mask=rgba.split()[3])
    return rgb


def save_pdf(pdf, folder, name):
    path = os.path.join(folder, f'{name}.pdf')
    pdf.output(path)
    return path


def add_df_to_pdf(pdf, df, header, font_size=10, line_height_x=2.5):
    df = df.astype(str)
    columns = df.columns
    records = df.to_records(index=False)
    if header:
        header = df.columns.str.upper()
    else:
        header = []

    pdf.set_font("Times", size=font_size)
    line_height = pdf.font_size * line_height_x
    col_width = pdf.epw / len(df.columns)
    for datum in header:
        pdf.multi_cell(col_width, line_height, datum, border=1,
                       ln=3, max_line_height=pdf.font_size)
    pdf.ln(line_height)

    for row in records:
        for datum in row:
            pdf.multi_cell(col_width, line_height, datum, border=1,
                           ln=3, max_line_height=pdf.font_size)
        pdf.ln(line_height)
    return pdf


def add_lineSpace_to_pdf(pdf, y, ln):
    pdf.ln(ln)
    y += 5+ln
    return pdf, y


def add_date_to_pdf(pdf):
    pdf.set_font('helvetica', '', 12)
    _today = str(pd.Timestamp.today()).split('.')[0]
    today_w = pdf.get_string_width(_today)+6
    doc_w = pdf.w
    pdf.set_x(doc_w-today_w)
    pdf.cell(today_w, txt=_today, ln=True)
    return pdf


def add_title_to_pdf(pdf, title):
    pdf.set_font('helvetica', '', 22)
    pdf.cell(w=pdf.get_string_width(title)+6,
             txt=title, ln=True)
    return pdf



import pandas as pd
from functools import lru_cache
from pandas.util import hash_pandas_object
from hashlib import sha256
import math
import numpy as np


def _index_marks(nrows, chunk_size):
    return range(chunk_size, math.ceil(nrows / chunk_size) * chunk_size, chunk_size)


def split_df(dfm, chunk_size):
    indices = _index_marks(dfm.shape[0], chunk_size)
    chunks = np.split(dfm, indices)
    return chunks


class HashableDataFrame(pd.DataFrame):
    def __init__(self, obj):
        super().__init__(obj)

    def __hash__(self):
        hash_value = sha256(hash_pandas_object(self, index=True).values)
        hash_value = hash(hash_value.hexdigest())
        return hash_value

    def __eq__(self, other):
        return self.equals(other)


class HashableSeries(pd.Series):
    def __init__(self, obj):
        super().__init__(obj)

    def __hash__(self):
        hash_value = sha256(hash_pandas_object(self, index=True).values)
        hash_value = hash(hash_value.hexdigest())
        return hash_value

    def __eq__(self, other):
        return self.equals(other)


def make_hashable_pandas_object(obj):
    if type(obj) is pd.DataFrame:
        return HashableDataFrame(obj)
    elif type(obj) is pd.Series:
        return HashableSeries(obj)
    else:
        raise Exception(f'obj must be of type DataFrame or Series, '
                        f'but is {type(obj)}')



gmail = ""
gmail_pw = ""

fred_key = ''

binance_key = ''
binance_secret = ''

census_bureau_key = ''

quandl_key = ''

twitter_key = ''
twitter_secret_key = ''
twitter_token = ''
twitter_token_secret = ''

eia_key = ''

newsapi_key = ''



from me_main_libs import *
import copy
import psutil


def create_logger(name):
    from loguru import logger
    logger.remove()
    logger_ = copy.deepcopy(logger)
    logger_.add(name,
                format='{time:YYYY-MM-DD HH:mm:ss.SSS} | {level: <8} | {name}:{function}:{line} - {message}', colorize=False)
    return logger_


def create_logger_2(folder, name):
    from loguru import logger
    logger.remove()
    logger_ = copy.deepcopy(logger)
    path = os.path.join(folder, f'{name}.log')
    logger_.add(path,
                format='{time:YYYY-MM-DD HH:mm:ss.SSS} | {level: <8} | {name}:{function}:{line} - {message}', colorize=False)
    return logger_


def create_parquet(df, folder, name):
    path = os.path.join(folder, f'{name}.parquet')
    df.to_parquet(path)
    return path


def create_html_from_df(df, folder, name, index=False, header=False, render_links=True):
    path = os.path.join(folder, f'{name}.html')
    df.to_html(path, index=index, header=header,
               render_links=render_links)
    return path


def create_png_from_plt(fig, folder, name):
    path = os.path.join(folder, f'{name}.png')
    fig.savefig(path)
    return path


def create_txt(data, folder, name):
    _file = os.path.join(folder, f'{name}.txt')
    with open(_file, 'w') as f:
        for line in data:
            f.write(str(line))
    return _file


def create_json(data, folder, name):
    _file = os.path.join(folder, f'{name}.json')
    with open(_file, "w") as f:
        json.dump(data, f)
    return _file


def read_parquet(folder, name):
    _file = os.path.join(folder, f'{name}.parquet')
    data = pd.read_parquet(_file)
    return data


def read_txt(folder, name, readlines=False):
    _file = os.path.join(folder, f'{name}.txt')
    with open(_file, 'r') as f:
        if readlines:
            x = f.readlines()
        else:
            x = f.read()
    return x


def read_log(folder, name, readlines=False):
    _file = os.path.join(folder, f'{name}.log')
    with open(_file, 'r') as f:
        if readlines:
            x = f.readlines()
        else:
            x = f.read()
    return x


def read_json(folder, name):
    _file = os.path.join(folder, f'{name}.json')
    with open(_file) as f:
        data = json.load(f)
    return data


def update_txt(data, folder, name):
    _file = os.path.join(folder, f'{name}.txt')
    with open(_file, 'a') as f:
        for line in data:
            f.write(str(line))
    return _file


def erase_txt(folder, name):
    _file = os.path.join(folder, f'{name}.txt')
    open(_file, "w").close()


def erase_log(folder, name):
    _file = os.path.join(folder, f'{name}.log')
    open(_file, "w").close()


def erase_json(folder, name):
    _file = os.path.join(folder, f'{name}.json')
    open(_file, "w").close()


def delete_files(files):
    for x in files:
        os.remove(x)


def delete_folder(folder):
    pass


def convert_py_to_ipynb(file):
    out_file = f'{file[:-3]}.ipynb'
    command = f'ipynb-py-convert {file} {out_file}'

    p = psutil.Popen(command, shell=True)
    while p.is_running():
        time.sleep(1)
    return out_file


def convert_ipynb_to_html(file):
    out_file = f'{file[:-6]}.html'
    command = f'jupyter nbconvert --to html {file}'

    p = psutil.Popen(command, shell=True)
    while p.is_running():
        time.sleep(1)
    return out_file


def convert_py_to_html(file):
    ipynb = convert_py_to_ipynb(file)
    html = convert_ipynb_to_html(ipynb)
    return html



from me_main_libs import *
from functools import lru_cache


def read_ts_inverted(ts_df):
    ts_df = ts_df*-1
    ts_df = ts_df+ts_df.min()*-1
    ts_df.columns = [x+'_inverted' for x in ts_df.columns]
    return ts_df


def read_ts_negative_zone(ts_df):
    ts_df = ts_df.copy()
    ts_df[ts_df > 0] = 0
    ts_df.columns = [x+'_negative_zone' for x in ts_df.columns]
    return ts_df


def read_ts_positive_zone(ts_df):
    ts_df = ts_df.copy()
    ts_df[ts_df < 0] = 0
    ts_df.columns = [x+'_positive_zone' for x in ts_df.columns]
    return ts_df


def read_ts_drawdown(ts_df):
    data = {}
    for x in ts_df.columns:
        data[x+'_drawdown'] = _read_ts_drawdown(ts_df[x])
    data = pd.concat(data, axis=1)
    return data


def read_ts_drawdown_rolling(ts_df, window):
    '''
    takes too much time
    '''
    data = pd.DataFrame()
    for x in ts_df.columns:
        col = x + f'_drawdown_roll_{window}'
        data[col] = _read_ts_rolling_drawdown(ts_df[x], window)
    return data


def _read_ts_drawdown(ts):
    if ts.min() < 0:
        ts = ts+abs(ts.min())
    ts = ts.dropna()
    running_max = np.maximum.accumulate(ts)
    running_max[running_max < 1] = 1
    return (ts)/running_max - 1


def _read_ts_rolling_drawdown(ts, window):
    data = ts.rolling(window, min_periods=window)
    data = [_read_ts_drawdown(
        x).tail(1) for x in data]
    data = pd.concat(data)
    return data


def read_ts_performance(ts_series, name):
    returns = ts_series.pct_change().dropna()
    returns.name = name+'_ret'

    cum_return = (1+returns).cumprod()
    cum_return.name = name+'_cret'

    drawdown = read_ts_drawdown(
        pd.DataFrame(cum_return)).squeeze()
    drawdown.name = name+'_dd'

    performance = {'cum_return': cum_return,
                   'drawdown': drawdown,
                   'returns': returns}
    return performance


def read_ts_performance_vs_spy(ts_series, name):
    ts1 = read_ts_performance(ts_series, name)
    spy = read_ts_performance(spy_adjClose(), 'spy')

    returns = pd.concat([ts1['returns'], spy['returns']],
                        axis=1).fillna(method='ffill').dropna()
    cum_returns = pd.concat([ts1['cum_return'], spy['cum_return']], axis=1).fillna(
        method='ffill').dropna()
    drawdowns = pd.concat([ts1['drawdown'], spy['drawdown']],
                          axis=1).fillna(method='ffill').dropna()

    performance = {'cum_returns': cum_returns,
                   'drawdowns': drawdowns,
                   'returns': returns}
    return performance


@lru_cache(maxsize=1)
def spy_adjClose():
    from yahoo_fin import stock_info
    data = stock_info.get_data('SPY')['adjclose']
    return data



from me_main_libs import *


def compute_performance_basic_portfolio(prices, portfolio_weights=None):
    returns = prices.pct_change().dropna()
    n_assets = len(prices.columns)
    if not portfolio_weights:
        portfolio_weights = n_assets * [1 / n_assets]

    portfolio_returns = pd.Series(np.dot(portfolio_weights, returns.T),
                                  index=returns.index)
    return portfolio_returns


def compute_needed_gain_to_restore_loss(percentage_loss=10):
    '''
    http://shurwest.com/wp-content/uploads/2013/08/The-Math-of-Gains-Losses.pdf
    As the loss gets bigger, the needed gain to restore the loss increases at an increasing rate. 
    '''
    result = (1/(1-percentage_loss/100))-1
    result = int(result*100)
    # result = f'{result}%'
    return result


def plot_needed_gain_to_restore_loss_relationship():
    results = {}
    for i in range(5, 80, 5):
        key = f'{-i}%'
        results[key] = compute_needed_gain_to_restore_loss(i)

    fig, ax = plt.subplots(1, 1)
    pd.Series(results).plot.bar(ax=ax)
    ax.set_ylabel('needed gain to restore loss (%)')
    ax.set_xlabel('percentage loss')
    plt.close()
    return fig



import signal
import os
import psutil


def find_procs_by_name(name):
    "Return a list of processes matching 'name'."
    ls = []
    for p in psutil.process_iter(['name']):
        if p.info['name'] == name:
            ls.append(p)
    return ls


def kill_proc_children(sig=signal.SIGTERM,
                       timeout=None, on_terminate=None):
    """Kill a process tree (including grandchildren) with signal
    "sig".
    "on_terminate", if specified, is a callback function which is
    called as soon as a child terminates.
    """
    try:
        parent = psutil.Process(os.getpid())
        children = parent.children(recursive=True)
        for p in children:
            p.kill()
    except:
        return 1
    return 0


def kill_proc_tree(pid, sig=signal.SIGTERM, include_parent=False,
                   timeout=None, on_terminate=None):
    """Kill a process tree (including grandchildren) with signal
    "sig" and return a (gone, still_alive) tuple.
    "on_terminate", if specified, is a callback function which is
    called as soon as a child terminates.
    """
    try:
        assert pid != os.getpid(), "won't kill myself"
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        if include_parent:
            children.append(parent)
        for p in children:
            try:
                p.send_signal(sig)
            except psutil.NoSuchProcess:
                pass
        gone, alive = psutil.wait_procs(children, timeout=timeout,
                                        callback=on_terminate)
        for p in children:
            p.kill()
    except:
        return 1
    return 0


def read_processes_running():
    processes = [(p.pid, p.info) for p in psutil.process_iter(
        ['name', 'status']) if p.info['status'] == psutil.STATUS_RUNNING]
    return processes



from me_main_libs import *
from pandas.tseries.offsets import BDay


def read_difference_between_dates(start_date, end_date, output_unit='hours'):
    '''
    output_unit: years,days,hours,minutes,seconds,microseconds
    '''
    time_diff = end_date-start_date
    time_diff = time_diff.total_seconds()

    if output_unit == 'years':
        seconds = 31536000
    if output_unit == 'days':
        seconds = 86400
    if output_unit == 'hours':
        seconds = 3600
    if output_unit == 'minutes':
        seconds = 60
    if output_unit == 'seconds':
        return time_diff.seconds
    if output_unit == 'microseconds':
        return time_diff.microseconds

    time_diff = divmod(time_diff, seconds)[0]
    return time_diff


def read_timestamp_to_integer(timestamp, output_unit='seconds'):
    '''
    output_unit: seconds,microseconds
    '''
    timestamp = timestamp.value
    if output_unit == 'seconds':
        return int(timestamp/1000000000)
    if output_unit == 'microseconds':
        return int(timestamp/1000000)


def read_business_days(start_date, end_date):
    dates = pd.date_range(start_date, end_date,
                          freq=BDay())
    return dates



import setuptools

setuptools.setup(
    name="me_helpers_lib",
    version="0.0.1",
    packages=setuptools.find_packages(),
    install_requires=['me_main_libs', 'bokeh',
                      'pandas', 'numpy',
                      'psutil', 'Pillow', 'PyPDF2',
                      'redis-decorator', 'redis'
                      ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)





import smtplib
from email import encoders
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from . import me_secrets


def send_mail(subject, body=None, to=me_secrets.gmail, files_to_send=[]):
    FROM = me_secrets.gmail
    TO = to

    msg = MIMEMultipart("alternative")
    msg["From"] = FROM
    msg["To"] = TO
    msg["Subject"] = subject

    if body:
        text_part = MIMEText(body, "plain")
        msg.attach(text_part)

    for file_ in files_to_send:
        with open(file_, "rb") as f:
            data = f.read()
            attach_part = MIMEBase("application", "octet-stream")
            attach_part.set_payload(data)
        encoders.encode_base64(attach_part)
        filename = file_.split('\\')[-1]
        attach_part.add_header("Content-Disposition",
                               f"attachment; filename= {filename}")
        msg.attach(attach_part)

    server = smtplib.SMTP("smtp.gmail.com", 587)
    server.starttls()
    server.login(me_secrets.gmail, me_secrets.gmail_pw)
    server.sendmail(FROM, TO, msg.as_string())
    server.quit()



from me_main_libs import *
from matplotlib import cm
from numpy import linspace
from functools import wraps
import signal
import psutil
import copy


# TODO: move to me_files
def create_logger(name):
    from loguru import logger
    logger.remove()
    logger_ = copy.deepcopy(logger)
    logger_.add(name,
                format='{time:YYYY-MM-DD HH:mm:ss.SSS} | {level: <8} | {name}:{function}:{line} - {message}')
    return logger_


def time_func(func):
    """
    use it as decorator: @time_func at the top of function
    """
    @ wraps(func)
    def measure_time(*args, **kwargs):
        t1 = time.time()
        result = func(*args, **kwargs)
        t2 = time.time()
        print(f"@timefn: {func.__name__} took {t2 - t1} seconds")
        return result
    return measure_time


def time_func_log(logger_var):
    def time_func(func):
        """
        use it as decorator: @time_func_log at the top of function
        """
        @ wraps(func)
        def measure_time(self, *args, **kwargs):
            t1 = time.time()
            result = func(self, *args, **kwargs)
            t2 = time.time()
            getattr(self, logger_var).debug(
                f"{func.__name__} took {t2 - t1} seconds")
            return result
        return measure_time
    return time_func


def make_colors(number):
    print('\n\n\n use create_colors !!!\n\n\n')


def create_colors(number):
    start = 0.0
    stop = 1.0
    number_of_lines = 9
    cm_subsection = linspace(start, stop, number)

    # here we choose coolwarm, check dir(cm) and
    # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
    # for more
    colors = [cm.coolwarm(x) for x in cm_subsection]
    return colors


def create_requests_session(headers_str):
    headers = {}
    for x in headers_str.split('\n'):
        y = x.split(':', maxsplit=1)
        headers[y[0].strip()] = y[1].strip()

    session = requests.Session()
    session.headers.clear()
    session.headers.update(headers)
    return session


def create_requests_headers(headers_str):
    headers = {}
    for x in headers_str.split('\n'):
        y = x.split(':', maxsplit=1)
        headers[y[0].strip()] = y[1].strip()
    return headers
